import { EventEmitter } from 'events';
import { Database } from 'better-sqlite3';

/**
 * Options for initializing the Migrator.
 */
interface MigratorOptions {
    /**
     * SQLite database instance
     */
    db: Database;
    /**
     * Directory containing migration files
     */
    migrationsDir: string;
    /**
     * Name of the migrations table (default: schema_migrations)
     */
    migrationsTable?: string;
    /**
     * Name of the migrations lock table (default: schema_migrations_lock)
     */
    migrationsLockTable?: string;
    /**
     * File extensions to look for when loading migrations (default: ['ts', 'js'])
     * Note: .d.ts files are always ignored
     */
    fileExtensions?: string[];
}
/**
 * A record of applied migrations as stored in the database.
 */
interface MigrationRecord {
    /**
     * Name of the migration
     */
    name: string;
    /**
     * ISO timestamp when the migration was executed
     */
    executed_at: string;
    /**
     * Batch number for the migration
     */
    batch: number;
}
/**
 * Result type returned by migration actions.
 */
interface MigrationResult {
    /**
     * Whether the migration operation was successful
     */
    success: boolean;
    /**
     * Error if the operation failed
     */
    error?: Error;
    /**
     * List of migrations that were applied/rolled back
     */
    appliedMigrations: string[];
}
/**
 * Represents the plan for pending migrations.
 */
interface MigrationPlan {
    /**
     * The next batch number that will be used
     */
    nextBatch: number;
    /**
     * List of migrations that will be applied
     */
    pendingMigrations: string[];
}
/**
 * Represents the status of migrations.
 */
interface MigrationStatus {
    /**
     * Current highest batch number
     */
    currentBatch: number;
    /**
     * Number of pending migrations
     */
    pending: number;
    /**
     * List of applied migrations
     */
    applied: MigrationRecord[];
}

/**
 * Migration provider for SQLite databases.
 * @example
 * const db = new Database('database.db');
 * const migrator = new Migrator({ db, migrationsDir: 'migrations' });
 * await migrator.apply();
 */
declare class Migrator extends EventEmitter {
    private db;
    private migrationsDir;
    private migrationsTable;
    private lockTable;
    private migrations;
    private initialized;
    private fileExtensions;
    constructor(options: MigratorOptions);
    /**
     * Ensures the `schema_migrations` and `schema_migrations_lock` tables exist,
     * and loads migration files from the migrations directory.
     */
    private init;
    /**
     * Initializes the `schema_migrations` and `schema_migrations_lock` tables.
     */
    private initTables;
    /**
     * Load migration files from the `migrationsDir`.
     * Migration files must export { up, down }.
     * The migrations are loaded in alphabetical order.
     */
    private loadMigrationsFromDirectory;
    /**
     * Acquire a lock to prevent concurrent migrations.
     * Throws an error if the lock is already held.
     */
    private acquireLock;
    /**
     * Release the migration lock.
     */
    private releaseLock;
    /**
     * Get the highest batch number.
     */
    private getCurrentBatch;
    /**
     * Insert a record for an applied migration.
     */
    private recordMigration;
    /**
     * Delete a record for a migration that is being rolled back.
     */
    private removeMigration;
    /**
     * Run SQL operations in a transaction.
     * @param fn The function to run in the transaction.
     */
    private runTransaction;
    /**
     * Apply all pending migrations in a single batch.
     * Returns the names of applied migrations.
     */
    apply(): Promise<MigrationResult>;
    /**
     * Roll back the most recent batch of migrations.
     * Returns the names of rolled back migrations.
     */
    rollback(): Promise<MigrationResult>;
    /**
     * Get the status of migrations.
     * Returns:
     * - currentBatch: the highest batch number applied
     * - pending: number of migrations not yet applied
     * - applied: list of all applied migrations
     */
    status(): Promise<MigrationStatus>;
    /**
     * Plan the pending migrations without applying them.
     * Returns the next batch number and the list of pending migration names in order.
     */
    plan(): Promise<MigrationPlan>;
}

export { type MigrationPlan, type MigrationRecord, type MigrationResult, type MigrationStatus, Migrator, type MigratorOptions };
