// src/index.ts
import path from "path";
import { promises as fs } from "fs";
import { EventEmitter } from "events";

// src/errors.ts
var MigrationError = class extends Error {
  constructor(message, cause) {
    super(message);
    this.cause = cause;
    this.name = "MigrationError";
  }
};
var MigrationFileError = class extends MigrationError {
  constructor(message, cause) {
    super(message, cause);
    this.name = "MigrationFileError";
  }
};
var MigrationLockError = class extends MigrationError {
  constructor(message, cause) {
    super(message, cause);
    this.name = "MigrationLockError";
  }
};
var MigrationExecutionError = class extends MigrationError {
  constructor(message, cause) {
    super(message, cause);
    this.name = "MigrationExecutionError";
  }
};

// src/index.ts
var Migrator = class extends EventEmitter {
  constructor(options) {
    super();
    this.migrations = [];
    this.initialized = false;
    this.db = options.db;
    this.migrationsDir = options.migrationsDir;
    this.migrationsTable = options.migrationsTable ?? "schema_migrations";
    this.lockTable = options.migrationsLockTable ?? "schema_migrations_lock";
    this.fileExtensions = options.fileExtensions ?? ["ts", "js"];
  }
  /**
   * Ensures the `schema_migrations` and `schema_migrations_lock` tables exist,
   * and loads migration files from the migrations directory.
   */
  async init() {
    if (this.initialized) {
      return;
    }
    await this.initTables();
    await this.loadMigrationsFromDirectory();
    this.initialized = true;
  }
  /**
   * Initializes the `schema_migrations` and `schema_migrations_lock` tables.
   */
  async initTables() {
    try {
      this.db.exec(`
        CREATE TABLE IF NOT EXISTS ${this.migrationsTable} (
          name TEXT PRIMARY KEY,
          executed_at TEXT NOT NULL,   -- ISO string
          batch INTEGER NOT NULL
        )
      `);
      this.db.exec(`
        CREATE TABLE IF NOT EXISTS ${this.lockTable} (
          id INTEGER PRIMARY KEY CHECK (id = 1),
          locked INTEGER NOT NULL DEFAULT 0
        )
      `);
      this.db.exec(`
        INSERT OR IGNORE INTO ${this.lockTable} (id, locked) VALUES (1, 0)
      `);
    } catch (err) {
      throw new MigrationError("Failed to initialize migrator", err);
    }
  }
  /**
   * Load migration files from the `migrationsDir`.
   * Migration files must export { up, down }.
   * The migrations are loaded in alphabetical order.
   */
  async loadMigrationsFromDirectory() {
    try {
      const entries = await fs.readdir(this.migrationsDir);
      const migrationFiles = entries.filter((file) => {
        if (file.endsWith(".d.ts")) {
          return false;
        }
        return this.fileExtensions.some((ext) => file.endsWith(`.${ext}`));
      }).sort();
      const loadedMigrations = [];
      for (const file of migrationFiles) {
        const fullPath = path.join(this.migrationsDir, file);
        let imported;
        try {
          imported = await import(fullPath);
        } catch (err) {
          throw new MigrationFileError(
            `Error loading migration "${file}": ${String(err)}`,
            err
          );
        }
        const { up, down } = imported;
        if (typeof up !== "function" || typeof down !== "function") {
          throw new MigrationFileError(
            `Migration "${file}" must export "up" and "down" functions.`
          );
        }
        loadedMigrations.push({ name: file, up, down });
      }
      this.migrations = loadedMigrations;
    } catch (err) {
      throw err instanceof MigrationFileError ? err : new MigrationFileError("Failed to load migrations", err);
    }
  }
  /**
   * Acquire a lock to prevent concurrent migrations.
   * Throws an error if the lock is already held.
   */
  acquireLock() {
    const transaction = this.db.transaction(() => {
      const row = this.db.prepare(`SELECT locked FROM ${this.lockTable} WHERE id = 1`).get();
      if (row.locked === 1) {
        throw new MigrationLockError("Migration lock already held by another process.");
      }
      this.db.prepare(`UPDATE ${this.lockTable} SET locked = 1 WHERE id = 1`).run();
    });
    try {
      transaction();
    } catch (err) {
      if (err instanceof MigrationLockError) {
        throw err;
      }
      throw new MigrationLockError("Failed to acquire migration lock", err);
    }
  }
  /**
   * Release the migration lock.
   */
  releaseLock() {
    try {
      this.db.prepare(`UPDATE ${this.lockTable} SET locked = 0 WHERE id = 1`).run();
    } catch (err) {
      throw new MigrationLockError("Failed to release migration lock", err);
    }
  }
  /**
   * Get the highest batch number.
   */
  getCurrentBatch() {
    const row = this.db.prepare(`SELECT MAX(batch) as batch FROM ${this.migrationsTable}`).get();
    return row?.batch ?? 0;
  }
  /**
   * Insert a record for an applied migration.
   */
  recordMigration(name, batch) {
    const executedAt = (/* @__PURE__ */ new Date()).toISOString();
    this.db.prepare(
      `
        INSERT INTO ${this.migrationsTable} (name, executed_at, batch)
        VALUES (?, ?, ?)
      `
    ).run(name, executedAt, batch);
  }
  /**
   * Delete a record for a migration that is being rolled back.
   */
  removeMigration(name, batch) {
    this.db.prepare(
      `
        DELETE FROM ${this.migrationsTable}
        WHERE name = ? AND batch = ?
      `
    ).run(name, batch);
  }
  /**
   * Run SQL operations in a transaction.
   * @param fn The function to run in the transaction.
   */
  runTransaction(fn) {
    const transaction = this.db.transaction(fn);
    transaction();
  }
  /**
   * Apply all pending migrations in a single batch.
   * Returns the names of applied migrations.
   */
  async apply() {
    try {
      await this.init();
    } catch (err) {
      return {
        success: false,
        error: err,
        appliedMigrations: []
      };
    }
    try {
      this.acquireLock();
    } catch (err) {
      return {
        success: false,
        error: err,
        appliedMigrations: []
      };
    }
    const appliedMigrations = [];
    try {
      const currentBatch = this.getCurrentBatch();
      const nextBatch = currentBatch + 1;
      const appliedNames = new Set(
        this.db.prepare(`SELECT name FROM ${this.migrationsTable}`).all().map((row) => row.name)
      );
      const pendingMigrations = this.migrations.filter((m) => !appliedNames.has(m.name));
      if (pendingMigrations.length === 0) {
        return { success: true, appliedMigrations };
      }
      this.runTransaction(() => {
        for (const migration of pendingMigrations) {
          try {
            migration.up(this.db);
            this.recordMigration(migration.name, nextBatch);
            appliedMigrations.push(migration.name);
            this.emit("migration:applied", migration.name, nextBatch);
          } catch (err) {
            throw new MigrationExecutionError(
              `Failed to rollback migration "${migration.name}"`,
              err
            );
          }
        }
      });
      return { success: true, appliedMigrations };
    } catch (error) {
      const err = error instanceof MigrationExecutionError ? error : new MigrationExecutionError("Migration failed", error);
      return { success: false, error: err, appliedMigrations: [] };
    } finally {
      this.releaseLock();
    }
  }
  /**
   * Roll back the most recent batch of migrations.
   * Returns the names of rolled back migrations.
   */
  async rollback() {
    try {
      await this.init();
    } catch (err) {
      return {
        success: false,
        error: err,
        appliedMigrations: []
      };
    }
    try {
      this.acquireLock();
    } catch (err) {
      return {
        success: false,
        error: err,
        appliedMigrations: []
      };
    }
    const appliedMigrations = [];
    try {
      const currentBatch = this.getCurrentBatch();
      if (currentBatch === 0) {
        return { success: true, appliedMigrations };
      }
      const rows = this.db.prepare(
        `
          SELECT name
          FROM ${this.migrationsTable}
          WHERE batch = ?
          ORDER BY name DESC
        `
      ).all(currentBatch);
      if (rows.length === 0) {
        return { success: true, appliedMigrations };
      }
      this.runTransaction(() => {
        for (const row of rows) {
          const migration = this.migrations.find((m) => m.name === row.name);
          if (!migration) {
            throw new MigrationFileError(`Migration "${row.name}" not found.`);
          }
          try {
            migration.down(this.db);
            this.removeMigration(migration.name, currentBatch);
            appliedMigrations.push(migration.name);
            this.emit("migration:rollback", migration.name, currentBatch);
          } catch (err) {
            throw new MigrationExecutionError(
              `Failed to rollback migration "${migration.name}"`,
              err
            );
          }
        }
      });
      return { success: true, appliedMigrations };
    } catch (error) {
      const err = error instanceof MigrationExecutionError ? error : new MigrationExecutionError("Rollback failed", error);
      return { success: false, error: err, appliedMigrations };
    } finally {
      this.releaseLock();
    }
  }
  /**
   * Get the status of migrations.
   * Returns:
   * - currentBatch: the highest batch number applied
   * - pending: number of migrations not yet applied
   * - applied: list of all applied migrations
   */
  async status() {
    await this.init();
    try {
      const currentBatch = this.getCurrentBatch();
      const rows = this.db.prepare(
        `
          SELECT name, executed_at, batch
          FROM ${this.migrationsTable}
          ORDER BY batch ASC, name ASC
        `
      ).all();
      const appliedNames = new Set(rows.map((r) => r.name));
      const pending = this.migrations.length - appliedNames.size;
      return {
        currentBatch,
        pending,
        applied: rows
      };
    } catch (err) {
      throw new MigrationError("Failed to get migration status", err);
    }
  }
  /**
   * Plan the pending migrations without applying them.
   * Returns the next batch number and the list of pending migration names in order.
   */
  async plan() {
    await this.init();
    try {
      const currentBatch = this.getCurrentBatch();
      const nextBatch = currentBatch + 1;
      const appliedNames = new Set(
        this.db.prepare(`SELECT name FROM ${this.migrationsTable}`).all().map((row) => row.name)
      );
      const pendingMigrations = this.migrations.filter((m) => !appliedNames.has(m.name)).map((m) => m.name);
      return {
        nextBatch,
        pendingMigrations
      };
    } catch (err) {
      throw new MigrationError("Failed to create migration plan", err);
    }
  }
};
export {
  Migrator
};
//# sourceMappingURL=index.js.map